// src/index.ts
import {resolve, dirname, relative} from "path";
import fs from "fs/promises";
import {createFilter} from "@rollup/pluginutils";
import {normalizePath} from "vite";
import {Project} from "ts-morph";

// src/utils.ts
function isNativeObj(value) {
  return Object.prototype.toString.call(value) === "[object Object]";
}
function mergeObjects(sourceObj, targetObj) {
  const loop = [
    {
      source: sourceObj,
      target: targetObj
    }
  ];
  while (loop.length) {
    const {source, target} = loop.pop();
    Object.keys(target).forEach((key) => {
      if (isNativeObj(target[key])) {
        if (!isNativeObj(source[key])) {
          source[key] = {};
        }
        loop.push({
          source: source[key],
          target: target[key]
        });
      } else if (Array.isArray(target[key])) {
        if (!Array.isArray(source[key])) {
          source[key] = [];
        }
        loop.push({
          source: source[key],
          target: target[key]
        });
      } else {
        source[key] = target[key];
      }
    });
  }
  return sourceObj;
}

// src/index.ts
var src_default = (options = {}) => {
  const {
    include = ["**/*.vue", "**/*.ts", "**/*.tsx"],
    exclude = "node_modules/**",
    root = process.cwd(),
    projectOptions = null,
    cleanVueFileName = false
  } = options;
  const filter = createFilter(include, exclude);
  const sourceFiles = [];
  const project = new Project(mergeObjects({
    compilerOptions: {
      declaration: true,
      emitDeclarationOnly: true,
      noEmitOnError: true
    },
    tsConfigFilePath: resolve(root, "tsconfig.json"),
    skipAddingFilesFromTsConfig: true
  }, projectOptions != null ? projectOptions : {}));
  return {
    name: "vite:dts",
    apply: "build",
    enforce: "post",
    transform(code, id) {
      if (!code || !filter(id))
        return null;
      if (/\.vue(\?.*type=script.*)$/.test(id)) {
        const filePath = resolve(root, normalizePath(id.split("?")[0]));
        sourceFiles.push(project.createSourceFile(filePath + (/lang.ts/.test(id) ? ".ts" : ".js"), code));
      } else if (/\.tsx?$/.test(id)) {
        const filePath = resolve(root, normalizePath(id));
        sourceFiles.push(project.addSourceFileAtPath(filePath));
      }
    },
    async generateBundle(outputOptions) {
      const declarationDir = outputOptions.file ? dirname(outputOptions.file) : outputOptions.dir;
      const diagnostics = project.getPreEmitDiagnostics();
      console.log(project.formatDiagnosticsWithColorAndContext(diagnostics));
      project.emitToMemory();
      for (const sourceFile of sourceFiles) {
        const emitOutput = sourceFile.getEmitOutput();
        for (const outputFile of emitOutput.getOutputFiles()) {
          const filePath = outputFile.getFilePath();
          const content = outputFile.getText();
          const targetPath = resolve(declarationDir, relative(root, cleanVueFileName ? filePath.replace(".vue.d.ts", ".d.ts") : filePath));
          await fs.mkdir(dirname(targetPath), {recursive: true});
          await fs.writeFile(targetPath, cleanVueFileName ? content.replace(/['"](.+)\.vue['"]/g, '"$1"') : content, "utf8");
        }
      }
    }
  };
};
export {
  src_default as default
};
